# TODO this should not be needed in the future
name_to_id_map = {
    'NOWPreprocessor': 'vlxgoxlv',
    'NOWQdrantIndexer16': '11yxatzh',
    'NOWElasticIndexer': 'mdtuyaaj',
    'NOWAutoCompleteExecutor2': 'w5w084h7',
}
